# Copyright 2022 VMware, Inc.
# SPDX-License-Identifier: BSD-2
import datetime
import pymisp
import unittest
from feed_manager import storage
from feed_manager import generator

from typing import Dict
from typing import List


UTC_NOW = datetime.datetime.utcnow()
UTC_TOMORROW = UTC_NOW + datetime.timedelta(days=1)
TEST_MD5_1 = "a" * 32
TEST_MD5_2 = "b" * 32


class InMemoryStorageLayer(storage.AbstractWriter):
    """In-Memory storage layer used for tests."""

    def __init__(self) -> None:
        """Constructor."""
        super().__init__()
        self.manifest = {}
        self.events = {}
        self.hashes = []

    def load_manifest(self) -> Dict:
        """Implement interface."""
        if not self.manifest:
            raise FileNotFoundError
        return self.manifest

    def load_event(self, event_uuid: str) -> Dict:
        """Implement interface."""
        try:
            return self.events[event_uuid]
        except KeyError:
            raise FileNotFoundError

    def load_hashes(self) -> List[List[str]]:
        """Implement interface."""
        return self.hashes

    def save_event(self, event_uuid: str, event_feed: Dict) -> None:
        """Implement interface."""
        self.events[event_uuid] = event_feed

    def save_manifest(self, manifest: Dict) -> None:
        """Implement interface."""
        self.manifest = manifest

    def save_hashes(self, attribute_hashes: List[List[str]]) -> None:
        """Implement interface."""
        self.hashes = attribute_hashes


class TestGenerator(unittest.TestCase):
    """Class to test the generator module."""

    def test__attributes__two_days(self):
        """Test adding two attributes in two different days."""
        storage_layer = InMemoryStorageLayer()
        feed_generator = generator.DailyFeedGenerator(storage_layer)
        feed_generator.add_attribute("md5", "a" * 32)
        feed_generator.flush()
        feed_generator.set_clock(UTC_TOMORROW)
        feed_generator.add_attribute("md5", "b" * 32)
        feed_generator.flush()
        self.assertEqual(len(storage_layer.manifest.keys()), 2)
        self.assertEqual(len(storage_layer.events), 2)
        self.assertEqual(len(storage_layer.hashes), 2)

    def test_attributes__flush_twice(self):
        """Test adding two attributes in one day and flushing twice."""
        storage_layer = InMemoryStorageLayer()
        feed_generator = generator.DailyFeedGenerator(storage_layer)
        feed_generator.add_attribute("md5", "a" * 32)
        feed_generator.flush()
        feed_generator.add_attribute("md5", "b" * 32)
        feed_generator.flush()
        self.assertEqual(len(storage_layer.manifest.keys()), 1)
        self.assertEqual(len(storage_layer.events), 1)
        self.assertEqual(len(storage_layer.hashes), 2)

    def test_attributes__flush_once(self):
        """Test adding two attributes in one day and flushing once."""
        storage_layer = InMemoryStorageLayer()
        feed_generator = generator.DailyFeedGenerator(storage_layer)
        feed_generator.add_attribute("md5", "a" * 32)
        feed_generator.add_attribute("md5", "b" * 32)
        feed_generator.flush()
        self.assertEqual(len(storage_layer.manifest.keys()), 1)
        self.assertEqual(len(storage_layer.events), 1)
        self.assertEqual(len(storage_layer.hashes), 2)

    def test_attributes__flush_once__miss(self):
        """Test adding two attributes in one day and flushing twice but once too late."""
        storage_layer = InMemoryStorageLayer()
        feed_generator = generator.DailyFeedGenerator(storage_layer)
        feed_generator.add_attribute("md5", "a" * 32)
        feed_generator.flush()
        feed_generator.add_attribute("md5", "b" * 32)
        self.assertEqual(len(storage_layer.manifest.keys()), 1)
        self.assertEqual(len(storage_layer.events), 1)
        self.assertEqual(len(storage_layer.hashes), 1)


class TestFeedUtils(unittest.TestCase):
    """Class to test the feed utils class."""

    def test_contains_attribute(self):
        """Test contain attribute."""
        misp_event = pymisp.MISPEvent()
        misp_event.add_attribute(type="md5", value="a" * 32)
        self.assertTrue(generator.FeedUtils.contains_attribute(misp_event, "md5", "a" * 32))
        self.assertFalse(generator.FeedUtils.contains_attribute(misp_event, "md5", "b" * 32))

    def test_contains_attribute_with_data(self):
        """Test contain attribute with data."""
        misp_event = pymisp.MISPEvent()
        misp_event.add_attribute(type="md5", value=TEST_MD5_1, data=b"data")
        self.assertFalse(generator.FeedUtils.contains_attribute(misp_event, "md5", TEST_MD5_1))
        self.assertFalse(generator.FeedUtils.contains_attribute(misp_event, "md5", TEST_MD5_2))
        self.assertTrue(
            generator.FeedUtils.contains_attribute(misp_event, "md5", TEST_MD5_1, data=b"data")
        )
        self.assertFalse(
            generator.FeedUtils.contains_attribute(
                misp_event, "md5", TEST_MD5_1, data=b"datadata"
            )
        )


if __name__ == "__main__":
    unittest.main()



# Copyright 2022 VMware, Inc.
# SPDX-License-Identifier: BSD-2
import ddt
import io
import unittest
from feed_manager import storage


@ddt.ddt
class TestStorage(unittest.TestCase):
    """Class to test the storage module."""

    @ddt.data(
        (io.TextIOWrapper(io.BytesIO(b"hash1,hash2"), encoding="utf-8"), [["hash1", "hash2"]]),
        (io.TextIOWrapper(io.BytesIO(b""), encoding="utf-8"), []),
        (io.TextIOWrapper(io.BytesIO(b" "), encoding="utf-8"), []),
        ([""], []),
        ([" "], []),
        (["hash1,hash2"], [["hash1", "hash2"]]),
        (["hash1,hash2", "hash3,hash4"], [["hash1", "hash2"], ["hash3", "hash4"]]),
    )
    def test_parse_csv(self, args):
        """Test parsing CSV files."""
        csv_data, expected_output = args
        output = storage.AbstractReader._parse_csv(csv_data)
        self.assertEqual(output, expected_output)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# Copyright 2022 VMware, Inc.
# SPDX-License-Identifier: BSD-2
import argparse
import collections
import datetime
import json
import sys

from feed_manager import storage
from feed_manager import consumer

from typing import Dict
from typing import List
from typing import Tuple


def get_date_range(start_date: datetime.date, end_date: datetime.date) -> List[datetime.date]:
    """Get a range of date objects with no gaps."""
    ranges = {
        start_date + datetime.timedelta(days=x) for x in range((end_date - start_date).days)
    }
    ranges.add(start_date)
    ranges.add(end_date)
    return sorted(ranges)


def compute_statistics(indicators: List[Dict]) -> Dict[datetime.date, Tuple[int, Dict]]:
    """Compute statistics from a list of indicators."""
    ret = {}
    attribute_uuid_to_tags = collections.defaultdict(set)
    date_to_attribute_uuids = collections.defaultdict(list)
    for indicator in indicators:
        attribute_uuid_to_tags[indicator["attribute_uuid"]].update(indicator["tags"])
        date_object = datetime.datetime.strptime(
            indicator["timestamp"], consumer.FeedParser.DEFAULT_FMT
        ).date()
        date_to_attribute_uuids[date_object].append(indicator["attribute_uuid"])
    try:
        min_date = min(date_to_attribute_uuids)
        max_date = max(date_to_attribute_uuids)
    except ValueError:
        return ret
    ret = {}
    for date_object in get_date_range(min_date, max_date):
        tag_to_count = collections.Counter()
        try:
            nb_attributes = len(date_to_attribute_uuids[date_object])
            for attribute_uuid in date_to_attribute_uuids[date_object]:
                for tag in attribute_uuid_to_tags[attribute_uuid]:
                    tag_to_count[tag] += 1
        except KeyError:
            nb_attributes = 0
        ret[date_object] = (nb_attributes, tag_to_count)
    return ret


def main():
    """Simple script to consume a daily feed."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-i",
        "--input-location",
        dest="input_location",
        type=str,
        default=None,
        required=True,
        help="location of the feed",
    )
    parser.add_argument(
        "-p",
        "--path",
        dest="path",
        type=str,
        default=None,
        help="optional additional path",
    )
    parser.add_argument(
        "-t",
        "--attribute-type",
        dest="attribute_type",
        type=str,
        default=None,
        help="the attribute type (e.g., sha1)",
    )
    parser.add_argument(
        "-g",
        "--galaxy-name",
        dest="galaxy_name",
        type=str,
        default=None,
        help="filter by MISP galaxy (e.g., malpedia)",
    )
    parser.add_argument(
        "-d",
        "--day-delta",
        dest="day_delta",
        type=int,
        default=7,
        help="the look back in terms of days",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        default=False,
        help="whether to be verbose and print attributes hashes",
    )
    args = parser.parse_args()

    storage_layer = storage.get_storage_layer(
        input_string=args.input_location,
        path=args.path,
        read_write=False,
    )
    feed_consumer = consumer.FeedConsumer(storage_layer)
    since_date_object = datetime.datetime.utcnow() - datetime.timedelta(days=args.day_delta)
    indicators = feed_consumer.get_items_since(
        date_object=since_date_object,
        attribute_type=args.attribute_type,
        galaxy_name=args.galaxy_name,
    )

    print(f"Fetching items since {since_date_object}")
    for indicator in indicators:
        print(json.dumps(indicator, indent=True))
    if args.verbose:
        print("Fetching all attribute hashes")
        for item in storage_layer.load_hashes():
            print(item[0], item[1])

    print("Computing statistics")
    try:
        statistics = compute_statistics(indicators)
    except KeyError:
        print("Not supported on telemetry feeds")
    else:
        for date_object, (nb_attributes, tag_to_count) in statistics.items():
            print(f"{date_object} - indicators: {nb_attributes}")
            for tag, count in sorted(tag_to_count.items()):
                print(f"\t{tag}: {count}")

    return 0


if __name__ == "__main__":
    sys.exit(main())



# Copyright 2022 VMware, Inc.
# SPDX-License-Identifier: BSD-2
import collections
import datetime
import itertools
import feed_manager

from typing import List
from typing import Generator
from typing import Iterable
from typing import Optional
from typing import Tuple
from typing import Union


try:
    import pymisp
except ImportError as ie:
    pymisp = None
    feed_manager.print_dependency_error_and_raise(ie)


class TagUtils:
    """Class with utility methods to handle tags."""

    # map object name to the attribute used to store object tags
    OBJECT_NAME_TO_ATTRIBUTE_TAG = {
        "file": "filename",
        "network-profile": "text",
        "sandbox-report": "score",
        "atp-report": "score",
    }

    @classmethod
    def __workaround_bug(cls, object_attribute: pymisp.MISPAttribute) -> pymisp.MISPAttribute:
        """Workaround the current pymisp bug where object attributes have mis-initialized tags."""
        if object_attribute.tags:
            return object_attribute
        if not hasattr(object_attribute, "AttributeTag"):
            return object_attribute
        tags = []
        try:
            for attribute_tag in object_attribute.AttributeTag:
                tag = attribute_tag["Tag"]
                tag_object = pymisp.MISPTag()
                tag_object.from_dict(**tag)
                tags.append(tag_object)
        except KeyError:
            pass
        was_edited = object_attribute.edited
        object_attribute.tags = tags
        if not was_edited:
            object_attribute.edited = False
        return object_attribute

    @classmethod
    def get_taggable_entity(
        cls,
        entity: Union[pymisp.MISPEvent, pymisp.MISPObject, pymisp.MISPAttribute],
    ) -> Union[pymisp.MISPEvent, pymisp.MISPAttribute]:
        """Get the taggable entity."""
        if isinstance(entity, pymisp.MISPEvent):
            return entity
        elif isinstance(entity, pymisp.MISPAttribute):
            return entity
        else:
            try:
                attribute_type = cls.OBJECT_NAME_TO_ATTRIBUTE_TAG[entity.name]
            except KeyError:
                raise ValueError(f"Can not process object '{entity.name}/{entity.uuid}'")
            try:
                return cls.__workaround_bug(entity.get_attributes_by_relation(attribute_type)[0])
            except IndexError:
                raise ValueError(f"Object '{entity.name}/{entity.uuid}' seems malformed'")

    @classmethod
    def validate_tag(cls, input_object: Union[pymisp.MISPTag, str]) -> pymisp.MISPTag:
        """Validate a tag (whether an object or a string) and return an object."""
        if isinstance(input_object, pymisp.MISPTag):
            return input_object
        else:
            return cls.create_tag(name=input_object)

    @classmethod
    def create_tag(cls, name: str, colour: Optional[str] = None) -> pymisp.MISPTag:
        """Create a tag."""
        tag = pymisp.MISPTag()
        tag.from_dict(
            name=name,
            colour=colour,
        )
        return tag

    @classmethod
    def add_tag(
        cls,
        entity: Union[pymisp.MISPAttribute, pymisp.MISPObject, pymisp.MISPEvent],
        tag: Union[str, pymisp.MISPTag],
    ):
        """Add a tag to a MISP entity, either an event, an object, or an attribute."""
        tag = cls.validate_tag(tag)
        taggable_entity = cls.get_taggable_entity(entity)
        taggable_entity.add_tag(tag)

    @classmethod
    def entity_contains_tag(
        cls,
        entity: Union[pymisp.MISPEvent, pymisp.MISPObject, pymisp.MISPAttribute],
        tag_name: str,
    ) -> bool:
        """Whether an entity is tagged with a given tag."""
        entity = cls.get_taggable_entity(entity)
        return any(tag.name == tag_name for tag in entity.tags)

    @classmethod
    def entity_contains_any_tags(
        cls,
        entity: Union[pymisp.MISPEvent, pymisp.MISPObject, pymisp.MISPAttribute],
        tag_names: List[str],
    ) -> bool:
        """Whether an entity is tagged with any of the given tags."""
        return any(cls.entity_contains_tag(entity, tag_name) for tag_name in tag_names)

    @classmethod
    def get_cluster_category_and_value(
        cls,
        tag: Union[str, pymisp.MISPTag],
    ) -> Tuple[Optional[str], Optional[str]]:
        """Decode a galaxy cluster tag into category and value."""
        if isinstance(tag, pymisp.MISPTag):
            tag = tag.name
        try:
            category, value = tag.split("=")
            return category, value.strip('"')
        except (KeyError, IndexError, ValueError):
            return None, None

    @classmethod
    def get_cluster_galaxy_and_value(
        cls,
        tag: Union[str, pymisp.MISPTag],
    ) -> Tuple[Optional[str], Optional[str]]:
        """Decode a galaxy cluster tag into category and value."""
        if isinstance(tag, pymisp.MISPTag):
            tag = tag.name
        category, value = cls.get_cluster_category_and_value(tag)
        if category is None and value is None:
            return None, None
        if not category.startswith("misp-galaxy:"):
            return None, None
        category = category.replace("misp-galaxy:", "")
        return category, value

    @classmethod
    def get_cluster_tag_value(
        cls,
        tag: Union[str, pymisp.MISPTag],
        category: str = None,
    ) -> Optional[str]:
        """Get the value of a galaxy cluster tag optionally filtering by category."""
        tag_category, tag_value = cls.get_cluster_category_and_value(tag)
        if category and tag_category:
            return tag_value if tag_category.startswith(category) else None
        else:
            return tag_value

    @classmethod
    def iter_tags(cls, event: pymisp.MISPEvent) -> Generator[pymisp.MISPTag, None, None]:
        """Iterate over all tags."""
        for tag in event.tags:
            yield tag
        for entity in itertools.chain(event.objects, event.attributes):
            try:
                taggable_entity = cls.get_taggable_entity(entity)
                for tag in taggable_entity.tags:
                    yield tag
            except ValueError:
                pass


class SightingUtils:
    """Utility class to deal with sightings and timestamps."""

    DATETIME_FMT = "%Y-%m-%d %H:%M:%S"

    @classmethod
    def format_sighting(
        cls,
        sighting_object: pymisp.MISPSighting,
        fmt: Optional[str] = None,
    ) -> str:
        """Format a sighting into a string."""
        if not fmt:
            fmt = cls.DATETIME_FMT
        return cls.timestamp_to_date(cls.sighting_to_timestamp(sighting_object)).strftime(fmt)

    @classmethod
    def format_seen_times(
        cls,
        misp_object: pymisp.MISPObject,
        fmt: Optional[str] = None,
    ) -> str:
        """Format first and last seen"""
        if not fmt:
            fmt = cls.DATETIME_FMT
        return "first={}/last={}".format(
            misp_object.first_seen.strftime(fmt), misp_object.last_seen.strftime(fmt)
        )

    @classmethod
    def is_datetime_naive(cls, date_object: datetime.datetime) -> bool:
        """Return whether a datetime object is naive (not TZ info)."""
        return date_object.tzinfo is None or date_object.tzinfo.utcoffset(date_object) is None

    @classmethod
    def timestamp_to_date(cls, timestamp: int) -> datetime.datetime:
        """Convert a timestamp to datetime object."""
        return datetime.datetime.utcfromtimestamp(timestamp).replace(tzinfo=datetime.timezone.utc)

    @classmethod
    def sighting_to_timestamp(cls, sighting_object: pymisp.MISPSighting) -> int:
        """Get the timestamp of a sighting object."""
        return int(sighting_object.date_sighting)

    @classmethod
    def get_sightings(
        cls, entity: Union[pymisp.MISPAttribute, pymisp.MISPObject]
    ) -> List[pymisp.MISPSighting]:
        """Get the all the sightings from an entity."""
        return TagUtils.get_taggable_entity(entity).Sighting

    @classmethod
    def get_sightings_date(
        cls, entity: Union[pymisp.MISPAttribute, pymisp.MISPObject]
    ) -> List[datetime.datetime]:
        """Get all sightings as datetime objects."""
        return [
            cls.timestamp_to_date(cls.sighting_to_timestamp(x)) for x in cls.get_sightings(entity)
        ]

    @classmethod
    def get_sightings_timestamp(
        cls, entity: Union[pymisp.MISPAttribute, pymisp.MISPObject]
    ) -> List[int]:
        """Get all sightings as timestamps."""
        return [cls.sighting_to_timestamp(x) for x in cls.get_sightings(entity)]

    @classmethod
    def date_to_sightings(
        cls,
        date_objects: Iterable[datetime.datetime],
    ) -> List[pymisp.MISPSighting]:
        """Update sightings."""
        return [cls.create_sighting(x) for x in date_objects]

    @classmethod
    def create_sighting(
        cls,
        date_object: datetime.datetime,
        sighting_type: str = "0",
        sighting_source: Optional[str] = None,
    ) -> pymisp.MISPSighting:
        """Create a sighting."""
        sighting_timestamp = date_object.timestamp()
        kwargs = {
            "timestamp": sighting_timestamp,  # this is read by pymisp
            "date_sighting": sighting_timestamp,  # this is for consistency with returned sightings
            "type": sighting_type,
        }
        if sighting_source:
            kwargs["source"] = sighting_source
        sighting = pymisp.MISPSighting()
        sighting.from_dict(**kwargs)
        return sighting

    @classmethod
    def has_sighting(
        cls,
        entity: Union[pymisp.MISPAttribute, pymisp.MISPObject],
        sighting: pymisp.MISPSighting,
        compare_source: bool = False,
    ) -> bool:
        """Return whether the object has a sighting with this timestamp."""
        # convert current and existing sightings into timestamps
        sighting_ts = cls.sighting_to_timestamp(sighting)
        sightings_by_ts = collections.defaultdict(list)
        for x in cls.get_sightings(entity):
            sightings_by_ts[cls.sighting_to_timestamp(x)].append(x)
        # if it is not in the mapping, we have no match
        if sighting_ts not in sightings_by_ts:
            return False
        # if we have a match, but we are not comparing the source, then we have it already
        if not compare_source:
            return True
        # otherwise compare all the sources
        return any(x.source == sighting.source for x in sightings_by_ts.get(sighting_ts, []))

    @classmethod
    def iter_sightings(
        cls,
        event: pymisp.MISPEvent,
    ) -> Generator[pymisp.MISPSighting, None, None]:
        """Iterate over all sightings."""
        for entity in itertools.chain(event.objects, event.attributes):
            try:
                for sighting in cls.get_sightings_date(entity):
                    yield sighting
            except ValueError:
                pass

    @classmethod
    def update_object_seen_times(
        cls,
        misp_object: pymisp.MISPObject,
        date_objects: Iterable[datetime.datetime],
    ) -> pymisp.MISPObject:
        """Update seen times of an object given a list of date objects."""
        if any(cls.is_datetime_naive(x) for x in date_objects):
            raise ValueError("Parameter 'date_objects' must contain timezone aware objects only")
        first_seen = min(date_objects)
        if (
            not hasattr(misp_object, "first_seen")
            or not misp_object.first_seen
            or first_seen < misp_object.first_seen
        ):
            misp_object.first_seen = first_seen
            misp_object.edited = True
        last_seen = max(date_objects)
        if (
            not hasattr(misp_object, "last_seen")
            or not misp_object.last_seen
            or last_seen > misp_object.last_seen
        ):
            misp_object.last_seen = last_seen
            misp_object.edited = True
        return misp_object


class IndicatorTranslator:
    """Class that translate indicators or telemetry events to MISP objects."""

    DEFAULT_FILE_ATTRIBUTE_CATEGORY = "Payload delivery"
    DEFAULT_NET_ATTRIBUTE_CATEGORY = "Network activity"
    DEFAULT_FILE_NAME = "unknown"

    @classmethod
    def to_network_attribute(
        cls,
        network_indicator: str,
        tags: Optional[List[Union[str, pymisp.MISPTag]]] = None,
    ) -> pymisp.MISPAttribute:
        """Get a network indicator as attribute."""
        if "://" in network_indicator:
            attribute_type = "url"
        elif len(network_indicator.split(".")) == 4:
            if ":" in network_indicator:
                attribute_type = "ip-dst|port"
            else:
                attribute_type = "ip-dst"
        else:
            attribute_type = "domain"
        net_attribute = pymisp.MISPAttribute()
        net_attribute.from_dict(
            type=attribute_type,
            category=cls.DEFAULT_NET_ATTRIBUTE_CATEGORY,
            value=network_indicator,
        )
        for tag in tags or []:
            TagUtils.add_tag(net_attribute, tag)
        return net_attribute

    @classmethod
    def to_file_attribute(
        cls,
        file_hash,
        attribute_category: str = DEFAULT_FILE_ATTRIBUTE_CATEGORY,
        tags: Optional[List[Union[str, pymisp.MISPTag]]] = None,
    ) -> pymisp.MISPAttribute:
        """Get a file attribute (useful when only a single hash is available)."""
        attribute_type = feed_manager.get_hash_type(file_hash)
        if not attribute_type:
            raise ValueError(f"Invalid hash '{file_hash}'")
        file_attribute = pymisp.MISPAttribute()
        file_attribute.from_dict(
            type=attribute_type,
            category=attribute_category,
            value=file_hash,
        )
        for tag in tags or []:
            TagUtils.add_tag(file_attribute, tag)
        return file_attribute

    @classmethod
    def to_file_object(
        cls,
        file_md5: str,
        file_sha1: str,
        file_sha256: str,
        file_name: Optional[str] = None,
        size: Optional[int] = None,
        mime_type: Optional[str] = None,
        comment: Optional[str] = None,
        attribute_category: str = DEFAULT_FILE_ATTRIBUTE_CATEGORY,
        tags: Optional[List[Union[str, pymisp.MISPTag]]] = None,
    ) -> pymisp.MISPObject:
        """Get a file object (mandatory when having multiple hashes)."""
        file_object = pymisp.MISPObject(name="file")
        file_object.add_attribute(
            "filename",
            value=file_name or cls.DEFAULT_FILE_NAME,
            comment=comment,
            to_ids=False,
        )
        if file_md5:
            file_object.add_attribute(
                "md5",
                value=file_md5,
                category=attribute_category,
                to_ids=True,
            )
        if file_sha1:
            file_object.add_attribute(
                "sha1",
                value=file_sha1,
                category=attribute_category,
                to_ids=True,
            )
        if file_sha256:
            file_object.add_attribute(
                "sha256",
                value=file_sha256,
                category=attribute_category,
                to_ids=True,
            )
        if size:
            file_object.add_attribute(
                "size-in-bytes",
                value=size,
                category=attribute_category,
                to_ids=False,
            )
        if mime_type:
            file_object.add_attribute(
                "mimetype",
                value=mime_type,
                category=attribute_category,
                to_ids=False,
            )
        for tag in tags or []:
            TagUtils.add_tag(file_object, tag)
        return file_object



import abc
import csv
import logging
import json
import os
import pathlib
import requests
from google.cloud import storage as google_storage

from typing import Dict
from typing import List
from typing import Optional
from typing import TextIO
from typing import Union


def get_storage_layer(
    input_string: str,
    path: Optional[str] = None,
    read_write: Optional[bool] = False,
):
    """Utility method to get the storage layer given some options."""
    logger = logging.getLogger(__name__)
    logger.info(
        "Creating storage layer with input_string='%s',path='%s',read_write=%s",
        input_string,
        path,
        read_write,
    )

    # http readers start with http
    if input_string.startswith("http"):
        return ReadOnlyHttpStorage(base_url=input_string)

    # local storage require a directory
    if os.path.isdir(input_string):
        feed_path = os.path.join(input_string, path or "")
        pathlib.Path(feed_path).mkdir(parents=True, exist_ok=True)
        if read_write:
            return ReadWriteLocalStorage(local_dir=feed_path)
        else:
            return ReadOnlyLocalStorage(local_dir=feed_path)

    # google storage uses bucket:config_file notation
    try:
        bucket, config_file = input_string.split(":")
        if read_write:
            return ReadWriteGoogleStorage(
                config_file=config_file,
                bucket_name=bucket,
                path=path,
            )
        else:
            return ReadOnlyGoogleStorage(
                config_file=config_file,
                bucket_name=bucket,
                path=path,
            )
    except ValueError:
        pass

    raise RuntimeError("Could not decode the storage layer from '%s'" % input_string)


class AbstractReader(abc.ABC):
    """Abstract reader."""

    HASHES_FILENAME = "hashes.csv"

    MANIFEST_FILENAME = "manifest.json"

    @classmethod
    def _parse_csv(cls, csv_data: Union[TextIO, List]) -> List[List[str]]:
        """Parse a CSV file."""
        if not csv_data:
            return []
        # skip empty lines and skip lines with just whitespace
        return [[row[0], row[1]] for row in csv.reader(csv_data) if row and row[0].strip()]

    def __init__(self, path: Optional[str] = None) -> None:
        self._path = path or ""
        self._logger = logging.getLogger(__name__)

    def _get_object_path(self, object_name) -> str:
        """Return the object name."""
        return os.path.join(self._path, object_name)

    @abc.abstractmethod
    def load_manifest(self) -> Dict:
        """Load manifest."""

    @abc.abstractmethod
    def load_event(self, event_uuid: str) -> Dict:
        """Load event."""

    @abc.abstractmethod
    def load_hashes(self) -> List[List[str]]:
        """Load hashes."""


class AbstractWriter(AbstractReader, abc.ABC):
    """Abstract writer."""

    @abc.abstractmethod
    def save_event(self, event_uuid: str, event_feed: Dict) -> None:
        """Save event."""

    @abc.abstractmethod
    def save_manifest(self, manifest: Dict) -> None:
        """Save manifest."""

    @abc.abstractmethod
    def save_hashes(self, attribute_hashes: List[List[str]]) -> None:
        """Save hashes."""

    def append_hashes(self, attribute_hashes: List[List[str]]) -> None:
        """Append hashes."""
        if not attribute_hashes:
            return
        hashes = self.load_hashes()
        hashes.extend(attribute_hashes)
        self.save_hashes(hashes)


class ReadOnlyLocalStorage(AbstractReader):
    """Read only storage from local directory."""

    def __init__(self, local_dir: str) -> None:
        """Constructor."""
        super().__init__(path=local_dir)

    def load_manifest(self) -> Dict:
        """Implement interface."""
        with open(self._get_object_path(self.MANIFEST_FILENAME), "r") as f:
            return json.load(f)

    def load_event(self, event_uuid: str) -> Dict:
        """Implement interface."""
        with open(self._get_object_path(f"{event_uuid}.json"), "r") as f:
            return json.load(f)

    def load_hashes(self) -> List[List[str]]:
        """Load hashes."""
        try:
            with open(self._get_object_path(self.HASHES_FILENAME), "r") as csv_file:
                return self._parse_csv(csv_file)
        except FileNotFoundError:
            return []


class ReadOnlyHttpStorage(AbstractReader):
    """Consumer using a remote (HTTP) source."""

    DEFAULT_TIMEOUT = 60

    def __init__(self, base_url: str) -> None:
        """Constructor."""
        super().__init__(path=None)
        self._base_url = base_url.rstrip("/")

    def _get_object_path(self, object_name: str) -> str:
        """Override method."""
        return f"{self._base_url}/{object_name}"

    def _download_data(self, object_name: str) -> requests.Response:
        """Download data from HTTP."""
        response = requests.get(self._get_object_path(object_name), timeout=self.DEFAULT_TIMEOUT)
        if response.status_code == 404:
            raise FileNotFoundError
        else:
            return response

    def load_manifest(self) -> Dict:
        """Implement interface."""
        return self._download_data(self.MANIFEST_FILENAME).json()

    def load_event(self, event_uuid: str) -> Dict:
        """Implement interface."""
        return self._download_data(f"{event_uuid}.json").json()

    def load_hashes(self) -> List[List[str]]:
        """Implement interface."""
        try:
            return self._parse_csv(self._download_data(self.HASHES_FILENAME).text.split("\n"))
        except FileNotFoundError:
            return []


class ReadOnlyGoogleStorage(AbstractReader):
    """Consumer using Google Storage."""

    def __init__(self, config_file: str, bucket_name: str, path: Optional[str] = None) -> None:
        """Constructor."""
        super().__init__(path=path)
        self._client = google_storage.Client.from_service_account_json(
            json_credentials_path=config_file,
        )
        self._bucket = self._client.bucket(bucket_name)

    def _download_data(self, object_name: str) -> str:
        """Download data from the bucket."""
        try:
            return self._bucket.get_blob(self._get_object_path(object_name)).download_as_text()
        except AttributeError:
            raise FileNotFoundError

    def load_manifest(self) -> Dict:
        """Implement interface."""
        return json.loads(self._download_data(self.MANIFEST_FILENAME))

    def load_event(self, event_uuid: str) -> Dict:
        """Implement interface."""
        return json.loads(self._download_data(f"{event_uuid}.json"))

    def load_hashes(self) -> List[List[str]]:
        try:
            return self._parse_csv(self._download_data(self.HASHES_FILENAME).split("\n"))
        except FileNotFoundError:
            return []


##
# WRITERS
##


class ReadWriteLocalStorage(ReadOnlyLocalStorage, AbstractWriter):
    """Writer using local storage."""

    def save_event(self, event_uuid: str, event_feed: Dict) -> None:
        """Implement interface."""
        with open(self._get_object_path(f"{event_uuid}.json"), "w") as f:
            json.dump(event_feed, f, indent=True)

    def save_manifest(self, manifest: Dict) -> None:
        """Implement interface."""
        with open(self._get_object_path(self.MANIFEST_FILENAME), "w") as f:
            json.dump(manifest, f, indent=True)

    def save_hashes(self, attribute_hashes: List[List[str]]) -> None:
        """Implement interface."""
        with open(self._get_object_path(self.HASHES_FILENAME), "w") as f:
            for element in attribute_hashes:
                f.write(f"{element[0]},{element[1]}\n")

    def append_hashes(self, attribute_hashes: List[List[str]]) -> None:
        """Override method."""
        with open(self._get_object_path(self.HASHES_FILENAME), "a") as f:
            for element in attribute_hashes:
                f.write(f"{element[0]},{element[1]}\n")


class ReadWriteGoogleStorage(ReadOnlyGoogleStorage, AbstractWriter):
    """Writer to Google Storage buckets."""

    def _upload_data(self, object_name: str, data: str) -> None:
        """Upload data to the bucket."""
        blob = self._bucket.blob(self._get_object_path(object_name))
        blob.upload_from_string(data)

    def save_event(self, event_uuid: str, event_feed: Dict) -> None:
        """Implement interface."""
        self._upload_data(object_name=f"{event_uuid}.json", data=json.dumps(event_feed))

    def save_manifest(self, manifest: Dict) -> None:
        """Implement interface."""
        self._upload_data(object_name=self.MANIFEST_FILENAME, data=json.dumps(manifest))

    def save_hashes(self, attribute_hashes: List[List[str]]) -> None:
        """Implement interface."""
        data = [f"{element[0]},{element[1]}" for element in attribute_hashes]
        self._upload_data(object_name=self.HASHES_FILENAME, data="\n".join(data))



# Copyright 2022 VMware, Inc.
# SPDX-License-Identifier: BSD-2
import ddt
import unittest
from feed_manager import translator


@ddt.ddt
class TestTranslator(unittest.TestCase):
    """Class to test the translator module."""

    @ddt.data(
        ("test=value", None, "value"),
        ('misp-galaxy:malpedia="test"', None, "test"),
        ('misp-galaxy:malpedia="test"', "misp-galaxy:malpedia", "test"),
        ('misp-galaxy:malpedia="test"', "misp-galaxy:another", None),
        ('misp-galaxy:malpedia2="test2"', None, "test2"),
        (translator.TagUtils.create_tag('misp-galaxy:malpedia="test"'), None, "test"),
    )
    def test_get_cluster_tag_value(self, args):
        """Test 'get_cluster_tag_value'."""
        tag, category, expected_output = args
        output = translator.TagUtils.get_cluster_tag_value(tag, category)
        self.assertEqual(output, expected_output)

    @ddt.data(
        ('malpedia2="test2"', ("malpedia2", "test2")),
        ('misp-galaxy:malpedia="test"', ("misp-galaxy:malpedia", "test")),
        ('misp-galaxy:malpedia2="test2"', ("misp-galaxy:malpedia2", "test2")),
        (
            translator.TagUtils.create_tag('misp-galaxy:malpedia="test"'),
            ("misp-galaxy:malpedia", "test"),
        ),
    )
    def test_get_cluster_category_and_value(self, args):
        """Test 'get_cluster_category_and_value'."""
        tag, expected_output = args
        output = translator.TagUtils.get_cluster_category_and_value(tag)
        self.assertEqual(output, expected_output)

    @ddt.data(
        ('malpedia2="test2"', (None, None)),
        ('misp-galaxy:malpedia="test"', ("malpedia", "test")),
        ('misp-galaxy:malpedia2="test2"', ("malpedia2", "test2")),
        (translator.TagUtils.create_tag('misp-galaxy:malpedia="test"'), ("malpedia", "test")),
    )
    def test_get_cluster_galaxy_and_value(self, args):
        """Test 'get_cluster_galaxy_and_value'."""
        tag, expected_output = args
        output = translator.TagUtils.get_cluster_galaxy_and_value(tag)
        self.assertEqual(output, expected_output)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# Copyright 2022 VMware, Inc.
# SPDX-License-Identifier: BSD-2
import argparse
import sys

from typing import Any
from typing import Dict
from typing import List

import feed_manager
from feed_manager import generator
from feed_manager import translator
from feed_manager import storage

try:
    import pymisp
except ImportError as ie:
    pymisp = None
    feed_manager.print_dependency_error_and_raise(ie)


INDICATOR_MD5 = "a" * 32

INDICATOR_SHA1 = "b" * 40

INDICATOR_SHA256 = "c" * 64

INDICATOR_TAGS = [
    'misp-galaxy:malpedia="GootKit"',
    'misp-galaxy:threat-actor="Sofacy"',
]

TELEMETRY_ITEM = {
    "file.sha1": "a1f7670cd7da7e331db2d69f0855858985819873",
    "file.mime_type": "application/x-pe-app-32bit-i386",
    "analysis.label": "unclassified",
    "task.severity": "malicious",
    "file.llfile_type": "PeExeFile",
    "analysis.mitre_tactics": ["TA0005: Defense Evasion", "TA0007: Discovery"],
    "file.md5": "37840d4e937db0385b820d4019071540",
    "task.portal_url": (
        "https://user.lastline.com/portal#/analyst/task/"
        "30f48c17e9db002005baa7d440ca275a/overview"
    ),
    "file.magic": "Unknown",
    "task.uuid": "30f48c17e9db002005baa7d440ca275a",
    "source.origin": "API",
    "file.sha256": "492bfe8d2b1105ec4045f96913d38f98e30fe349ea50cc4aaa425ca289af2852",
    "customer.sector": "IT",
    "analysis.activities": [
        "Anomaly: AI detected possible malicious code reuse",
        "Evasion: Detecting the presence of AntiMalware Scan Interface (AMSI)",
        "Execution: Subject crash detected",
        "Signature: Potentially malicious application/program",
    ],
    "analysis.mitre_techniques": ["T1497: Virtualization/Sandbox Evasion"],
    "task.score": 70,
    "customer.region": "AMER",
    "utc_timestamp": 1663568636000,
    "file.size": 990720,
    "file.name": None,
    "research.tag.name": [],
    "research.tag.value": [],
}

SANDBOX_NAME = "Test Sandbox"


def from_telemetry_item_to_objects(item: Dict[str, Any]) -> List[pymisp.MISPObject]:
    """Convert a telemetry item into objects."""
    objects = []
    file_object = translator.IndicatorTranslator.to_file_object(
        file_md5=item["file.md5"],
        file_sha1=item["file.sha1"],
        file_sha256=item["file.sha256"],
        file_name=item.get("file.name"),
        mime_type=item.get("file.mime_type"),
        size=item.get("file.size"),
    )
    objects.append(file_object)
    sandbox_object = pymisp.MISPObject(name="sandbox-report")
    sandbox_object.add_attribute("score", item["task.score"])
    sandbox_object.add_attribute("saas-sandbox", SANDBOX_NAME)
    sandbox_object.add_attribute("permalink", item["task.portal_url"])
    sandbox_object.add_reference(
        referenced_uuid=file_object.uuid,
        relationship_type="report-of",
    )
    objects.append(sandbox_object)
    if item.get("analysis.activities"):
        sig_object = pymisp.MISPObject(name="sb-signature")
        for activity in item["analysis.activities"]:
            sig_object.add_attribute("signature", type="text", value=activity)
        sig_object.add_reference(
            referenced_uuid=sandbox_object.uuid,
            relationship_type="belongs-to",
        )
        objects.append(sig_object)
    return objects


def main():
    """Simple script to generate a daily feed."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-o",
        "--output-location",
        dest="output_location",
        type=str,
        default="./tmp/",
        help="the output location",
    )
    args = parser.parse_args()

    # Test the indicator feed
    misp_object = translator.IndicatorTranslator.to_file_object(
        file_md5=INDICATOR_MD5,
        file_sha1=INDICATOR_SHA1,
        file_sha256=INDICATOR_SHA256,
        tags=INDICATOR_TAGS,
    )
    feed_generator = generator.DailyFeedGenerator(
        storage_layer=storage.get_storage_layer(
            input_string=args.output_location,
            path="indicators",
            read_write=True,
        ),
        feed_properties=generator.FeedProperties(
            title="Test indicators feed",
        ),
    )
    feed_generator.add_object(misp_object)
    feed_generator.flush()
    print("Daily feed of indicators flushed")

    # Test the telemetry feed
    misp_objects = from_telemetry_item_to_objects(TELEMETRY_ITEM)
    feed_generator = generator.DailyFeedGenerator(
        storage_layer=storage.get_storage_layer(
            input_string=args.output_location,
            path="telemetry",
            read_write=True,
        ),
        feed_properties=generator.FeedProperties(
            title="Test telemetry feed",
        ),
    )
    for misp_object in misp_objects:
        feed_generator.add_object(misp_object)
    feed_generator.flush()
    print("Daily feed of telemetry objects flushed")

    return 0


if __name__ == "__main__":
    sys.exit(main())



# Copyright 2023 VMware, Inc.
# SPDX-License-Identifier: BSD-2
import ddt
import unittest

from unittest.mock import MagicMock

from feed_manager import consumer


@ddt.ddt
class TestFeedConsumer(unittest.TestCase):
    """Class to test the FeedConsumer class."""

    FEED_MANIFEST = {
        "UUID2023_06_03_00": {"timestamp": 1685750400},
        "UUID2023_06_04_00": {"timestamp": 1685836800},
        "UUID2023_06_05_00": {"timestamp": 1685923200},
        "UUID2023_06_05_04": {"timestamp": 1685930400},
    }

    @ddt.data(
        (1686009600, 1686009600),  # Tue Jun 06 2023 00:00:00 GMT+0000
        (1686002400, 1685923200),  # Mon Jun 05 2023 00:00:00 GMT+0000
        (1686048389, 1686009600),  # Tue Jun 06 2023 00:00:00 GMT+0000
        (1686002399, 1685923200),  # Mon Jun 05 2023 00:00:00 GMT+0000
    )
    def test_round_timestamp_to_day(self, args):
        input_timestamp, expected_timestamp = args
        returned_timestamp = consumer.FeedConsumer._round_timestamp_to_day(input_timestamp)
        self.assertEqual(expected_timestamp, returned_timestamp)

    @ddt.data(
        (1685750400, ["UUID2023_06_03_00"]),
        (1685836824, ["UUID2023_06_04_00"]),
        (1685955600, ["UUID2023_06_05_00", "UUID2023_06_05_04"]),
    )
    def test__get_event_uuids_on(self, args):
        timestamp, expected_uuids = args
        c = consumer.FeedConsumer(None)
        c._storage_layer = MagicMock()
        c._storage_layer.load_manifest.return_value = self.FEED_MANIFEST
        returned_uuids = c._get_event_uuids_on(timestamp)
        self.assertEqual(returned_uuids, expected_uuids)


if __name__ == "__main__":
    unittest.main()



# Copyright 2022 VMware, Inc.
# SPDX-License-Identifier: BSD-2
import abc
import base64
import collections
import datetime
import logging

from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import feed_manager
from feed_manager import translator

try:
    import pymisp
except ImportError as ie:
    pymisp = None
    feed_manager.print_dependency_error_and_raise(ie)


class FeedGenerationException(Exception):
    """Generic exception."""


class FeedEventNotFound(FeedGenerationException):
    """Exception raised when the feed has not the required event."""


FeedEventMetadata = collections.namedtuple(
    "FeedEventMetadata",
    [
        "event_uuid",
        "event_bucket",
    ],
)


class FeedProperties:
    """Class to represent the properties of a feed."""

    DEFAULT_ORG_NAME = "Organization"
    DEFAULT_ORG_UUID = "a6b0e426-8c54-4136-beb6-3c2bafac7c4b"
    DEFAULT_FEED_NAME = "Feed"

    @classmethod
    def create_organization(
        cls,
        name: str = DEFAULT_ORG_NAME,
        uuid: str = DEFAULT_ORG_UUID,
    ) -> pymisp.MISPOrganisation:
        organization = pymisp.MISPOrganisation()
        organization.from_dict(
            name=name,
            uuid=uuid,
        )
        return organization

    def __init__(
        self,
        analysis: pymisp.Analysis = pymisp.Analysis.initial,
        threat_level_id: pymisp.ThreatLevel = pymisp.ThreatLevel.low,
        title: str = DEFAULT_FEED_NAME,
        published: bool = False,
        tags: Optional[List[Union[pymisp.MISPTag, str]]] = None,
        organization: Optional[pymisp.MISPOrganisation] = None,
    ):
        if not organization:
            organization = self.create_organization()
        if not tags:
            tags = []
        self.organization = organization
        self.analysis = analysis
        self.tags = [translator.TagUtils.validate_tag(x) for x in tags]
        self.threat_level_id = threat_level_id
        self.published = published
        self.title = title


class FeedUtils:
    """Feed utilities."""

    @staticmethod
    def attribute_equals(attr1: pymisp.MISPAttribute, attr2: pymisp.MISPAttribute) -> bool:
        """Return whether two attributes are the same."""

        def _data_equals() -> bool:
            if attr1.data is None:
                return attr2.data is None
            elif attr2.data is None:
                return attr1.data is None
            else:
                return (
                    base64.b64encode(attr1.data.getvalue()).decode()
                    == base64.b64encode(attr2.data.getvalue()).decode()
                )

        return attr1.type == attr2.type and attr1.value == attr2.value and _data_equals()

    @staticmethod
    def tag_equals(tag1: pymisp.MISPTag, tag2: pymisp.MISPTag) -> bool:
        """Return whether two tags are the same."""
        return tag1.name == tag2.name

    @classmethod
    def object_equals(cls, obj1: pymisp.MISPObject, obj2: pymisp.MISPObject) -> bool:
        """Return whether two objects are the same."""
        obj1_attributes = sorted(obj1.attributes, key=lambda x: x.type)
        obj2_attributes = sorted(obj2.attributes, key=lambda x: x.type)
        if len(obj1_attributes) != len(obj2_attributes):
            return False
        return all(
            cls.attribute_equals(attr1, attr2)
            for attr1, attr2 in zip(obj1_attributes, obj2_attributes)
        )

    @classmethod
    def contains_attribute(
        cls,
        misp_event: pymisp.MISPEvent,
        attr_type: str,
        attr_value: str,
        **attr_data,
    ) -> bool:
        """Return whether the misp event contains a specific attribute."""
        fake_attribute = pymisp.MISPAttribute()
        fake_attribute.from_dict(
            type=attr_type,
            value=attr_value,
            **attr_data,
        )
        return any(cls.attribute_equals(fake_attribute, attr) for attr in misp_event.attributes)

    @classmethod
    def contains_object(
        cls, misp_event: pymisp.MISPEvent, misp_object: pymisp.MISPObject
    ) -> bool:
        """Return whether the misp event contains a specific object."""
        return any(cls.object_equals(obj, misp_object) for obj in misp_event.objects)

    @classmethod
    def contains_tag(cls, misp_event: pymisp.MISPEvent, misp_tag: pymisp.MISPTag) -> bool:
        """Return whether the misp event contains a specific tag."""
        return any(cls.tag_equals(tag, misp_tag) for tag in misp_event.tags)


class AbstractFeedGenerator(abc.ABC):
    """Abstract class for every feed generators."""

    def __init__(self, storage_layer):
        """Constructor."""
        self._logger = logging.getLogger(__name__)
        self._storage_layer = storage_layer
        self._manifest = {}
        self._hashes = []

    @abc.abstractmethod
    def add_object(self, misp_object: pymisp.MISPObject) -> bool:
        """Add object to the feed."""

    @abc.abstractmethod
    def add_tag(self, misp_tag: pymisp.MISPTag) -> bool:
        """Add tag to the feed."""

    @abc.abstractmethod
    def add_attribute(self, attr_type: str, attr_value: str, **attr_data) -> bool:
        """Add an attribute to the feed."""

    @abc.abstractmethod
    def flush(self) -> None:
        """Flush the feed."""

    def _flush_event(
        self, event: pymisp.MISPEvent, update_manifest: Optional[bool] = True
    ) -> None:
        """Flush the event."""
        if update_manifest:
            self._manifest.update(event.manifest)
            self._storage_layer.save_manifest(self._manifest)
        self._storage_layer.save_event(event.uuid, event.to_feed())
        self._storage_layer.append_hashes(self._hashes)
        self._hashes.clear()

    def _add_hash(self, event: pymisp.MISPEvent, attr_type: str, attr_value: str) -> None:
        """Take the attribute properties and add a hash."""
        fake_attribute = pymisp.MISPAttribute()
        fake_attribute.from_dict(type=attr_type, value=attr_value)
        for hash_value in fake_attribute.hash_values("md5"):
            self._hashes.append([hash_value, event.get("uuid")])


class PeriodicFeedGenerator(AbstractFeedGenerator, abc.ABC):
    """A periodic feed generator that needs to be specialized further."""

    @classmethod
    @abc.abstractmethod
    def get_bucket(cls, date_obj: datetime.datetime) -> str:
        """Return the periodic bucket given the provided date object."""

    @classmethod
    @abc.abstractmethod
    def parse_bucket(cls, date_str: str) -> datetime.datetime:
        """Given a bucket return the date time object."""

    @classmethod
    def get_bucket_from_timestamp(cls, timestamp: int) -> str:
        """Get the bucket given a timestamp."""
        return cls.get_bucket(datetime.datetime.utcfromtimestamp(timestamp))

    @classmethod
    def get_event_metadata(
        cls,
        manifest: Dict,
        event_bucket: Optional[str] = None,
    ) -> FeedEventMetadata:
        """Get the metadata related to the event."""
        if event_bucket:
            for event_uuid, event_json in manifest.items():
                parsed_bucket = cls.get_bucket_from_timestamp(event_json["timestamp"])
                if parsed_bucket == event_bucket:
                    return FeedEventMetadata(
                        event_uuid=event_uuid,
                        event_bucket=parsed_bucket,
                    )
            raise FeedEventNotFound
        else:
            dated_events = []
            for event_uuid, event_json in manifest.items():
                dated_events.append(
                    (
                        event_json["timestamp"],
                        event_uuid,
                        event_json["info"],
                    )
                )
            # Sort by date then by event name
            dated_events.sort(key=lambda k: (k[0], k[2], k[1]), reverse=True)
            return FeedEventMetadata(
                event_uuid=dated_events[0][1],
                event_bucket=cls.get_bucket_from_timestamp(dated_events[0][0]),
            )

    def set_clock(self, utc_now: Optional[datetime.datetime]) -> None:
        """Initialize the feed clock to the provided value or utcnow() otherwise."""
        if utc_now:
            self._event_date_callback = lambda: utc_now
        else:
            self._event_date_callback = lambda: datetime.datetime.utcnow()

    def __init__(
        self,
        storage_layer,
        feed_properties: Optional[FeedProperties] = None,
        date_override: Optional[datetime.datetime] = None,
    ):
        """Constructor."""
        super(PeriodicFeedGenerator, self).__init__(storage_layer)
        self._feed_properties = feed_properties or FeedProperties()
        # Let 'set_clock' initialize '_event_date_callback'
        self._event_date_callback = None
        # Set up the callback used to know the current date at which we are inserting items
        self.set_clock(date_override)

        # Load the manifest but create it in case it is empty
        self._logger.info("Loading feed manifest")
        try:
            self._manifest = self._storage_layer.load_manifest()
        except FileNotFoundError:
            self._logger.info("No manifest found; creating new manifest and first event")
            self._flush_event(self._create_event(self._get_current_bucket()))

        # Load the current event but load it if the existing manifest does not have it
        if date_override:
            event_bucket = self.get_bucket(date_override)
            self._logger.info(
                "Loading feed event with an overriden date/bucket: %s/%s",
                date_override,
                event_bucket,
            )
            try:
                event_metadata = self.get_event_metadata(
                    manifest=self._manifest,
                    event_bucket=event_bucket,
                )
            except FeedEventNotFound:
                self._logger.info(
                    "No feed event found; creating new event at the request date/bucket: %s/%s",
                    date_override,
                    event_bucket,
                )
                self._flush_event(self._create_event(event_bucket))
                event_metadata = self.get_event_metadata(
                    manifest=self._manifest,
                    event_bucket=event_bucket,
                )
        else:
            self._logger.info("Loading feed event with the latest date/bucket")
            event_metadata = self.get_event_metadata(self._manifest)
        self._current_event_bucket = event_metadata.event_bucket
        self._current_event = self._load_event(event_metadata.event_uuid)

    def _get_current_bucket(self) -> str:
        """Get the current bucket (truncated datetime object)."""
        return self.get_bucket(self._event_date_callback())

    def _update_event(self) -> None:
        """Update the current bucket if needed."""
        event_bucket = self._get_current_bucket()
        if self._current_event_bucket != event_bucket:
            self._logger.info(
                "New event bucket required (current=%s, requested=%s)",
                self._current_event_bucket,
                event_bucket,
            )
            # flush previous event without updating manifest
            self._flush_event(self._current_event, update_manifest=False)
            self._current_event_bucket = event_bucket
            self._current_event = self._create_event(event_bucket)
            self._flush_event(self._current_event)

    def _create_event(self, event_bucket: str) -> pymisp.MISPEvent:
        """Create an event in the given bucket."""
        event = pymisp.MISPEvent()
        event.from_dict(
            **{
                "id": str(len(self._manifest) + 1),
                "info": f"{self._feed_properties.title} ({event_bucket})",
                "date": event_bucket,
                "analysis": self._feed_properties.analysis.value,
                "threat_level_id": self._feed_properties.threat_level_id.value,
                "published": self._feed_properties.published,
                "timestamp": int(self.parse_bucket(event_bucket).timestamp()),
            }
        )
        for tag in self._feed_properties.tags:
            event.add_tag(tag)
        event.Orgc = self._feed_properties.organization
        return event

    def _load_event(self, event_uuid: str) -> pymisp.MISPEvent:
        """Load the event."""
        data = self._storage_layer.load_event(event_uuid)
        event = pymisp.MISPEvent()
        event.from_dict(**data["Event"])
        return event

    ##
    # PUBLIC METHODS
    ##

    def add_object(self, misp_object: pymisp.MISPObject) -> bool:
        """Implement interface."""
        self._update_event()
        if FeedUtils.contains_object(self._current_event, misp_object):
            return False
        self._current_event.add_object(misp_object)
        for attribute in misp_object.attributes:
            self._add_hash(self._current_event, attribute.type, attribute.value)
        return True

    def add_attribute(self, attr_type: str, attr_value: str, **attr_data) -> bool:
        """Implement interface."""
        self._update_event()
        if FeedUtils.contains_attribute(self._current_event, attr_type, attr_value, **attr_data):
            return False
        self._current_event.add_attribute(attr_type, attr_value, **attr_data)
        self._add_hash(self._current_event, attr_type, attr_value)
        return True

    def add_tag(self, misp_tag: pymisp.MISPTag) -> bool:
        """Implement interface."""
        self._update_event()
        if FeedUtils.contains_tag(self._current_event, misp_tag):
            return False
        self._current_event.add_tag(misp_tag)
        return True

    def flush(self) -> None:
        """Implement interface."""
        self._flush_event(self._current_event)


class DailyFeedGenerator(PeriodicFeedGenerator):
    """A feed generator that creates a different event every day."""

    BUCKET_FMT = "%Y-%m-%d"

    @classmethod
    def get_bucket(cls, date_obj: datetime.datetime) -> str:
        """Implement interface."""
        return date_obj.replace(
            hour=0,
            minute=0,
            second=0,
            microsecond=0,
        ).strftime(cls.BUCKET_FMT)

    @classmethod
    def parse_bucket(cls, date_str: str) -> datetime.datetime:
        """Implement interface"""
        return datetime.datetime.strptime(date_str, cls.BUCKET_FMT).replace(
            tzinfo=datetime.timezone.utc
        )
